"""
Tests for the sync client's API.

"""

import csv
from typing import Any, cast

import pytest

import akismet
from akismet import _common

pytestmark = [pytest.mark.anyio, pytest.mark.api, pytest.mark.async_client]


@pytest.mark.parametrize(
    "method", ["CONNECT", "DELETE", "HEAD", "OPTIONS", "PATCH", "PUT", "TRACE"]
)
async def test_unsupported_request_method(
    akismet_config: akismet.Config,
    akismet_async_client: akismet.AsyncClient,
    method: str,
):
    """
    Attempting to make a request with an unsupported method raises AkismetError.

    """
    with pytest.raises(akismet.AkismetError):
        await akismet_async_client._request(
            method,  # type: ignore
            _common._API_V11,
            _common._COMMENT_CHECK,
            {"api_key": akismet_config.key},
        )


async def test_verify_key_valid(akismet_async_client: akismet.AsyncClient):
    """
    verify_key() returns True when the config is valid.

    """
    assert await akismet_async_client.verify_key()


@pytest.mark.akismet_client(verify_key_response=False)
async def test_verify_key_invalid(akismet_async_client: akismet.AsyncClient):
    """
    verify_key() returns False when the config is invalid.

    """
    assert not await akismet_async_client.verify_key()


async def test_verify_key_valid_explicit(
    akismet_async_client: akismet.AsyncClient, akismet_config: akismet.Config
):
    """
    verify_key() returns True when the config is valid and explicitly passed in.

    """
    assert await akismet_async_client.verify_key(
        key=akismet_config.key, url=akismet_config.url
    )


@pytest.mark.akismet_client(verify_key_response=False)
async def test_verify_key_invalid_explicit(
    akismet_async_client: akismet.AsyncClient, akismet_config: akismet.Config
):
    """
    verify_key() returns False when the config is invalid and explicitly
    passed in.

    """
    assert not await akismet_async_client.verify_key(
        key=akismet_config.key, url=akismet_config.url
    )


@pytest.mark.akismet_fixed_response(response_text="invalid")
@pytest.mark.parametrize(
    ["method_name", "pass_args"],
    [
        ("comment_check", True),
        ("submit_ham", True),
        ("submit_spam", True),
        ("key_sites", False),
        ("usage_limit", False),
    ],
    ids=["comment_check", "submit_ham", "submit_spam", "key_sites", "usage_limit"],
)
async def test_request_with_invalid_key(
    akismet_async_client_fixed_response: akismet.AsyncClient,
    akismet_common_kwargs: dict,
    method_name: str,
    pass_args: bool,
):
    """
    The request methods other than verify_key() raise akismet.APIKeyError if called with
    an invalid API key/URL.

    """
    method = getattr(akismet_async_client_fixed_response, method_name)
    args = akismet_common_kwargs if pass_args else {}
    with pytest.raises(akismet.APIKeyError):
        await method(**args)


@pytest.mark.parametrize(
    "expected",
    [
        pytest.param(
            akismet.CheckResponse.HAM,
            marks=pytest.mark.akismet_client(
                comment_check_response=akismet.CheckResponse.HAM
            ),
            id="ham",
        ),
        pytest.param(
            akismet.CheckResponse.SPAM,
            marks=pytest.mark.akismet_client(
                comment_check_response=akismet.CheckResponse.SPAM
            ),
            id="spam",
        ),
        pytest.param(
            akismet.CheckResponse.DISCARD,
            marks=pytest.mark.akismet_client(
                comment_check_response=akismet.CheckResponse.DISCARD
            ),
            id="discard",
        ),
    ],
)
async def test_comment_check(
    akismet_async_client: akismet.AsyncClient,
    akismet_common_kwargs: dict,
    expected: akismet.CheckResponse,
):
    """
    comment_check() returns the expected value based on the Akismet API's response.

    """
    assert await akismet_async_client.comment_check(**akismet_common_kwargs) == expected


async def test_submit_ham(
    akismet_async_client: akismet.AsyncClient, akismet_common_kwargs: dict
):
    """
    submit_ham() returns True when Akismet accepts the submission.

    """
    assert await akismet_async_client.submit_ham(**akismet_common_kwargs)


async def test_submit_spam(
    akismet_async_client: akismet.AsyncClient, akismet_common_kwargs: dict
):
    """
    submit_spam() returns True when Akismet accepts the submission.

    """
    assert await akismet_async_client.submit_spam(**akismet_common_kwargs)


async def test_key_sites_json(akismet_async_client: akismet.AsyncClient):
    """
    key_sites() returns key usage information in JSON format by async default.

    """
    response_json = cast(dict[str, Any], await akismet_async_client.key_sites())
    for key in ["2022-09", "limit", "offset", "total"]:
        assert key in response_json
    sites = response_json["2022-09"]
    for site in sites:
        for key in [
            "site",
            "api_calls",
            "spam",
            "ham",
            "missed_spam",
            "false_positives",
            "is_revoked",
        ]:
            assert key in site


async def test_key_sites_csv(akismet_async_client: akismet.AsyncClient):
    """
    key_sites() returns key usage information in CSV format when requested.

    """
    first, *rest = (
        cast(str, await akismet_async_client.key_sites(result_format="csv"))
    ).splitlines()
    assert first.startswith("Active sites for")
    reader = csv.DictReader(rest)
    row = next(reader)
    assert set(row.keys()) == {
        "Site",
        "Total API Calls",
        "Spam",
        "Ham",
        "Missed Spam",
        "False Positives",
        "Is Revoked",
    }


async def test_usage_limit(akismet_async_client: akismet.AsyncClient):
    """
    usage_limit() returns the API usage statistics in JSON format.

    """
    response_json = await akismet_async_client.usage_limit()
    assert set(response_json.keys()) == {
        "limit",
        "usage",
        "percentage",
        "throttled",
    }



"""
Test configuration.

"""

import typing
from http import HTTPStatus
from unittest import mock

import httpx
import pytest

import akismet
from akismet import _test_clients

# pylint: disable=redefined-outer-name


@pytest.fixture
def anyio_backend() -> str:
    """
    Return the async backend to use for async tests.

    """
    return "asyncio"


@pytest.fixture
def akismet_common_kwargs() -> dict:
    """
    Return the common set of base arguments neded for most Akismet API calls.

    """
    return {"user_ip": "127.0.0.1"}


@pytest.fixture
def akismet_end_to_end_kwargs(akismet_common_kwargs: dict) -> dict:
    """
    Return the set of base arguments needed for end-to-end test calls.

    """
    return {"is_test": 1, **akismet_common_kwargs}


@pytest.fixture
def akismet_fixed_response_transport() -> (
    typing.Callable[[str, HTTPStatus, typing.Optional[dict]], httpx.MockTransport]
):
    """
    Factory function to generate an httpx transport with a fixed response.

    """

    def _fixed_response_transport(
        response_text: str = "true",
        status_code: HTTPStatus = HTTPStatus.OK,
        response_json: typing.Optional[dict] = None,
    ) -> httpx.MockTransport:
        """
        Return an ``httpx`` transport that produces a fixed response, for use
        in testing.

        The transport will return a response consisting of:

        * ``status_code`` (default 200)
        * ``response_json`` as the JSON content, if supplied
        * Otherwise ``response_text`` (default ``"true"``) as the response text

        """

        def _handler(
            request: httpx.Request,  # pylint: disable=unused-argument
        ) -> httpx.Response:
            """
            Mock transport handler which returns a controlled response.

            """
            response_kwargs = {"status_code": status_code, "content": response_text}
            if response_json is not None:
                del response_kwargs["content"]
                response_kwargs["json"] = response_json
            return httpx.Response(**response_kwargs)  # type: ignore

        return httpx.MockTransport(_handler)

    return _fixed_response_transport


@pytest.fixture
def akismet_sync_client_fixed_response(
    request: pytest.FixtureRequest,
    akismet_fixed_response_transport: typing.Callable[
        [str, HTTPStatus, typing.Optional[dict]], httpx.MockTransport
    ],
) -> akismet.SyncClient:
    """
    Return a (sync) Akismet test client with an HTTP client which produces a fixed
    response.

    """
    marker = request.node.get_closest_marker("akismet_fixed_response")
    return akismet.SyncClient(
        http_client=httpx.Client(
            transport=akismet_fixed_response_transport(**marker.kwargs)  # type: ignore
        )
    )


@pytest.fixture
def akismet_async_client_fixed_response(
    request: pytest.FixtureRequest,
    akismet_fixed_response_transport: typing.Callable[
        [str, HTTPStatus, typing.Optional[dict]], httpx.MockTransport
    ],
) -> akismet.AsyncClient:
    """
    Return an (async) Akismet test client with an HTTP client which produces a fixed
    response.

    """
    marker = request.node.get_closest_marker("akismet_fixed_response")
    return akismet.AsyncClient(
        http_client=httpx.AsyncClient(
            transport=akismet_fixed_response_transport(**marker.kwargs)  # type: ignore
        )
    )


@pytest.fixture
def akismet_sync_http_client_exception(request: pytest.FixtureRequest) -> httpx.Client:
    """
    Return a (sync) HTTP client which always raises an exception.

    """
    marker = request.node.get_closest_marker("akismet_exception_response")
    exception_class = marker.kwargs.get("exception_class", Exception)
    error_message = marker.kwargs.get("error_message", "Error!")

    return mock.Mock(
        spec_set=httpx.Client,
        get=mock.Mock(side_effect=exception_class(error_message)),
        post=mock.Mock(side_effect=exception_class(error_message)),
    )


@pytest.fixture
def akismet_async_http_client_exception(
    request: pytest.FixtureRequest,
) -> httpx.AsyncClient:
    """
    Return an (async) HTTP client which always raises an exception.

    """
    marker = request.node.get_closest_marker("akismet_exception_response")
    exception_class = marker.kwargs.get("exception_class", Exception)
    error_message = marker.kwargs.get("error_message", "Error!")

    return mock.AsyncMock(
        spec_set=httpx.AsyncClient,
        get=mock.Mock(side_effect=exception_class(error_message)),
        post=mock.Mock(side_effect=exception_class(error_message)),
    )


@pytest.fixture
def akismet_sync_client_exception(
    akismet_sync_http_client_exception,
) -> akismet.SyncClient:
    """
    Return a (sync) Akismet test client whose HTTP client always raises an exception.

    """
    return akismet.SyncClient(http_client=akismet_sync_http_client_exception)


@pytest.fixture
def akismet_async_client_exception(
    akismet_async_http_client_exception,
) -> akismet.AsyncClient:
    """
    Return an (async) Akismet test client whose HTTP client always raises an exception.

    """
    return akismet.AsyncClient(http_client=akismet_async_http_client_exception)


@pytest.fixture
def akismet_config() -> akismet.Config:
    """
    Return an Akismet config object.

    """
    return akismet.Config(key=_test_clients._TEST_KEY, url=_test_clients._TEST_URL)


@pytest.fixture
def akismet_bad_config() -> akismet.Config:
    """
    Return an Akismet config object with invalid values.

    """
    return akismet.Config(key="INVALID_TEST_KEY", url="http://example.com")



"""
Common definitions used by both the sync and async Akismet implementations.

"""

# SPDX-License-Identifier: BSD-3-Clause

import enum
import os
import sys
import textwrap
from collections.abc import Mapping
from importlib.metadata import version
from typing import Literal, NamedTuple, NoReturn, TypedDict, cast

import httpx

from . import _exceptions

# Private constants.
# -------------------------------------------------------------------------------

_API_URL = "https://rest.akismet.com"
_API_V11 = "1.1"
_API_V12 = "1.2"
_COMMENT_CHECK = "comment-check"
_KEY_SITES = "key-sites"
_REQUEST_METHODS = Literal["GET", "POST"]  # pylint: disable=invalid-name
_SUBMISSION_RESPONSE = "Thanks for making the web a better place."
_SUBMIT_HAM = "submit-ham"
_SUBMIT_SPAM = "submit-spam"
_USAGE_LIMIT = "usage-limit"
_VERIFY_KEY = "verify-key"

_KEY_ENV_VAR = "PYTHON_AKISMET_API_KEY"
_URL_ENV_VAR = "PYTHON_AKISMET_BLOG_URL"

_TIMEOUT = float(os.getenv("PYTHON_AKISMET_TIMEOUT", "1.0"))


# Public constants.
# -------------------------------------------------------------------------------

USER_AGENT = (
    f"akismet.py/{version('akismet')} | Python/"
    f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
)


# Public classes.
# -------------------------------------------------------------------------------


class AkismetArguments(TypedDict, total=False):
    """
    A :class:`~typing.TypedDict` representing the optional keyword arguments accepted by
    the comment-check, submit-ham, and submit-spam Akismet API operations.

    """

    blog_charset: str
    blog_lang: str
    comment_author: str
    comment_author_email: str
    comment_author_url: str
    comment_content: str
    comment_context: str
    comment_date_gmt: str
    comment_post_modified_gmt: str
    comment_type: str
    honeypot_field_name: str
    is_test: bool
    permalink: str
    recheck_reason: str
    referrer: str
    user_agent: str
    user_role: str


class CheckResponse(enum.IntEnum):
    """
    Possible response values from an Akismet content check, including the
    possibility of the "discard" response, modeled as an :class:`enum.IntEnum`.

    """

    HAM = 0
    SPAM = 1
    DISCARD = 2


class Config(NamedTuple):
    """
    A :func:`~collections.namedtuple` representing Akismet configuration, consisting
    of a key and a URL.

    You only need to use this if you're manually configuring an Akismet API client
    rather than letting the configuration be read automatically from environment
    variables.

    """

    key: str
    url: str


# Private helper functions.
# -------------------------------------------------------------------------------


# Functions which throw errors for various situations.
# -------------------------------------------------------------------------------


def _configuration_error(config: Config) -> NoReturn:
    """
    Raise an appropriate exception for invalid configuration.

    """
    raise _exceptions.APIKeyError(
        textwrap.dedent(
            f"""
            Akismet API key and/or blog URL were invalid.

            Found API key: {config.key}
            Found URL: {config.url}
            """
        )
    )


def _protocol_error(operation: str, response: httpx.Response) -> NoReturn:
    """
    Raise an appropriate exception for unexpected API responses.

    """
    raise _exceptions.ProtocolError(
        textwrap.dedent(
            f"""
        Received unexpected or non-standard response from Akismet API.

        API operation was: {operation}
        API response received was: {response.text}
        Debug header value was: {response.headers.get('X-akismet-debug-help', None)}
        """
        )
    )


# Functions which help autodiscover/autofill configuration.
# -------------------------------------------------------------------------------


def _get_async_http_client() -> httpx.AsyncClient:
    """
    Return an asynchronous HTTP client for interacting with the Akismet API.

    """
    return httpx.AsyncClient(headers={"User-Agent": USER_AGENT}, timeout=_TIMEOUT)


def _get_sync_http_client() -> httpx.Client:
    """
    Return a synchronous HTTP client for interacting with the Akismet API.

    """
    return httpx.Client(headers={"User-Agent": USER_AGENT}, timeout=_TIMEOUT)


def _try_discover_config() -> Config:
    """
    Attempt to discover and return an Akismet configuration from the environment.

    :raises akismet.ConfigurationError: When either or both of the API key and
       URL are missing, or if the URL does not begin with ``"http://"`` or
       ``https://``.

    """
    key = os.getenv(_KEY_ENV_VAR, None)
    url = os.getenv(_URL_ENV_VAR, None)

    if key is None or url is None:
        raise _exceptions.ConfigurationError(
            textwrap.dedent(
                f"""
        Could not find full Akismet configuration.

        Found API key: {key}
        Found blog URL: {url}
        """
            )
        )

    if not url.startswith(("http://", "https://")):
        raise _exceptions.ConfigurationError(
            textwrap.dedent(
                f"""
            Invalid Akismet site URL specified: {url}

            Akismet requires the full URL including the leading 'http://' or 'https://'.
            """
            )
        )
    return Config(key=key, url=url)


# Functions which help process Akismet requests and responses.
# -------------------------------------------------------------------------------


def _handle_akismet_response(endpoint: str, response: httpx.Response) -> httpx.Response:
    """
    Check the response to see if it indicates an invalid key.

    """
    # It's possible to construct a client without performing up-front API key
    # validation, in which case the responses will all have text "invalid". So we check
    # for that and raise an exception when it's detected.
    if endpoint != _VERIFY_KEY and response.text == "invalid":
        raise _exceptions.APIKeyError("Akismet API key and/or site URL are invalid.")
    return response


def _handle_check_response(response: httpx.Response) -> CheckResponse:
    """
    Return the correct result for a response from the comment-check endpoint.

    """
    if response.text == "true":
        if response.headers.get("X-akismet-pro-tip", "") == "discard":
            return CheckResponse.DISCARD
        return CheckResponse.SPAM
    if response.text == "false":
        return CheckResponse.HAM
    _protocol_error(_COMMENT_CHECK, response)


def _handle_submit_response(endpoint: str, response: httpx.Response) -> bool:
    """
    Proces the response from a submit (ham/spam) request.

    """
    if response.text == _SUBMISSION_RESPONSE:
        return True
    _protocol_error(endpoint, response)


def _handle_verify_key_response(response: httpx.Response) -> bool:
    """
    Handle the response from a verify_key() request.

    """
    if response.text == "valid":
        return True
    if response.text == "invalid":
        return False
    _protocol_error(_VERIFY_KEY, response)


def _prepare_post_kwargs(kwargs: Mapping, endpoint: str) -> AkismetArguments:
    """
    Verify that the provided set of keyword arguments is valid for an Akismet POST
    request, returning them if they are or raising UnknownArgumentError if they aren't.

    """
    if unknown_args := [
        k
        for k in kwargs
        if k not in AkismetArguments.__optional_keys__  # pylint: disable=no-member
    ]:
        raise _exceptions.UnknownArgumentError(
            f"Received unknown argument(s) for Akismet operation {endpoint}: "
            f"{', '.join(unknown_args)}"
        )
    return cast(AkismetArguments, kwargs)


def _prepare_request(
    method: _REQUEST_METHODS, api_version: str, endpoint: str, data: dict
) -> tuple[str, dict]:
    """
    From the raw arguments passed to _request(), prepare the correct argument set to
    pass to the HTTP client and return them.

    """
    if method not in ("GET", "POST"):
        raise _exceptions.AkismetError(
            f"Unrecognized request method attempted: {method}."
        )
    request_kwarg = "data" if method == "POST" else "params"
    return f"{_API_URL}/{api_version}/{endpoint}", {request_kwarg: data}



"""
Tests for the async client's constructor behavior.

"""

import pytest

import akismet
from akismet import _common

pytestmark = [pytest.mark.anyio, pytest.mark.async_client, pytest.mark.constructors]


async def test_construct_config_explicit(
    akismet_config: akismet.Config, akismet_async_class: type[akismet.AsyncClient]
):
    """
    Passing explicit config to the async default constructor uses that config.

    """
    config = akismet.Config(key="invalid-test-key", url=akismet_config.url)
    async with akismet_async_class(config=config) as client:
        assert client._config == config


async def test_construct_config_alternate_constructor_explicit(
    akismet_config: akismet.Config, akismet_async_class: type[akismet.AsyncClient]
):
    """
    Passing explicit config to the alternate constructor uses that config.

    """
    config = akismet.Config(key="other-invalid-test-key", url=akismet_config.url)
    client = await akismet_async_class.validated_client(config=config)
    assert client._config == config


async def test_construct_config_from_env(
    akismet_config: akismet.Config, akismet_async_class: type[akismet.AsyncClient]
):
    """
    Instantiating via the async default constructor, without passing explicit config,
    reads the config from the environment.

    """
    async with akismet_async_class() as client:
        assert client._config == akismet_config


async def test_construct_alternate_constructor_config_from_env(
    akismet_config: akismet.Config, akismet_async_class: type[akismet.AsyncClient]
):
    """
    Instantiating via the alternate constructor, without passing explicit config, reads
    the config from the environment.

    """
    client = await akismet_async_class.validated_client()
    assert client._config == akismet_config


async def test_construct_config_valid(akismet_async_class: type[akismet.AsyncClient]):
    """
    With a valid configuration, constructing a client succeeds.

    """
    await akismet_async_class.validated_client()


@pytest.mark.akismet_client(verify_key_response=False)
async def test_construct_config_invalid_key(
    akismet_async_class: type[akismet.AsyncClient],
):
    """
    With an invalid API key, constructing a client raises an APIKeyError.

    """
    with pytest.raises(akismet.APIKeyError):
        await akismet_async_class.validated_client()


async def test_construct_config_valid_context_manager(
    akismet_async_class: type[akismet.AsyncClient],
):
    """
    With a valid configuration, constructing a client as a context manager succeeds.

    """
    async with akismet_async_class():
        pass


@pytest.mark.akismet_client(verify_key_response=False)
async def test_construct_config_invalid_key_context_manager(
    akismet_async_class: type[akismet.AsyncClient],
):
    """
    With an invalid API key, constructing a client as a context manager raises an
    APIKeyError.

    """
    with pytest.raises(akismet.APIKeyError):
        async with akismet_async_class():
            pass


async def test_construct_config_valid_explicit(
    akismet_config: akismet.Config, akismet_async_class: type[akismet.AsyncClient]
):
    """
    With an explicit valid configuration, constructing a client succeeds.

    """
    await akismet_async_class.validated_client(config=akismet_config)


@pytest.mark.akismet_client(verify_key_response=False)
async def test_construct_config_invalid_key_explicit(
    akismet_config: akismet.Config, akismet_async_class: type[akismet.AsyncClient]
):
    """
    With an explicit invalid API key, constructing a client raises an APIKeyError.

    """
    with pytest.raises(akismet.APIKeyError):
        await akismet_async_class.validated_client(config=akismet_config)


async def test_construct_config_bad_url(
    monkeypatch: pytest.MonkeyPatch, akismet_async_class: type[akismet.AsyncClient]
):
    """
    With an invalid URL, constructing a client raises a ConfigurationError.

    """
    monkeypatch.setenv(_common._URL_ENV_VAR, "ftp://example.com")
    with pytest.raises(akismet.ConfigurationError):
        await akismet_async_class.validated_client()


async def test_construct_config_missing_key(
    monkeypatch: pytest.MonkeyPatch,
    akismet_async_class: type[akismet.AsyncClient],
):
    """
    Without an API key present, constructing a client raises a ConfigurationError.

    """
    monkeypatch.delenv(_common._KEY_ENV_VAR)
    with pytest.raises(akismet.ConfigurationError):
        await akismet_async_class.validated_client()


async def test_construct_config_missing_url(
    monkeypatch: pytest.MonkeyPatch,
    akismet_async_class: type[akismet.AsyncClient],
):
    """
    Without a registered site URL present, constructing a client raises a
    ConfigurationError.

    """
    monkeypatch.delenv(_common._URL_ENV_VAR)
    with pytest.raises(akismet.ConfigurationError):
        await akismet_async_class.validated_client()


async def test_construct_config_missing_all(
    monkeypatch: pytest.MonkeyPatch,
    akismet_async_class: type[akismet.AsyncClient],
):
    """
    Without any config present, constructing a client raises a ConfigurationError.

    """
    monkeypatch.delenv(_common._KEY_ENV_VAR)
    monkeypatch.delenv(_common._URL_ENV_VAR)
    with pytest.raises(akismet.ConfigurationError):
        await akismet_async_class.validated_client()


async def test_construct_default_client():
    """
    Constructing a client without an explicit HTTP client uses the async default HTTP
    client.

    """
    client = akismet.AsyncClient()
    http_client = client._http_client
    assert "user-agent" in http_client.headers
    assert http_client.headers["user-agent"] == _common.USER_AGENT



"""
Pytest plugin supplying markers and fixtures to assist in testing code which uses the
Akismet clients.

"""

import pytest

from . import AsyncClient, CheckResponse, SyncClient, TestAsyncClient, TestSyncClient

# pylint: disable=redefined-outer-name


def _get_mark_arguments(request: pytest.FixtureRequest) -> tuple[CheckResponse, bool]:
    """
    Collect and return the arguments for construction of an Akismet test client, from
    pytest mark configuration if set or from defaults if not.

    """
    comment_check_response = CheckResponse.SPAM
    verify_key_response = True
    if marker := request.node.get_closest_marker("akismet_client"):
        comment_check_response = marker.kwargs.get(
            "comment_check_response", comment_check_response
        )
        verify_key_response = marker.kwargs.get(
            "verify_key_response", verify_key_response
        )
    return comment_check_response, verify_key_response


@pytest.fixture
def akismet_sync_class(request: pytest.FixtureRequest) -> type[TestSyncClient]:
    """
    A synchronous Akismet test client class.

    Use this fixture if you need the actual class itself (for example, to register with
    dependency injection tool which will instantiate it for you). If you just want an a
    client instance to use directly, use the akismet_sync_client fixture instead.

    By default, this will return a class whose verify_key() method always returns True,
    and whose comment_check() method always returns akismet.CheckResponse.SPAM. To
    change this, use the pytest mark akismet_client on your test, and pass either or
    both of the following keyword arguments:

    * comment_check_response: an akismet.CheckResponse which will be used as the return
      value of the comment_check() method.

    * verify_key_response: a bool which will be used as the return value of the
      verify_key() method.

    """
    _comment_check_response, _verify_key_response = _get_mark_arguments(request)

    class _SyncClient(TestSyncClient):
        """
        Test Akismet client.

        """

        comment_check_response = _comment_check_response
        verify_key_response = _verify_key_response

    return _SyncClient


@pytest.fixture
def akismet_async_class(request: pytest.FixtureRequest) -> type[TestAsyncClient]:
    """
    An asynchronous Akismet test client class.

    Use this fixture if you need the actual class itself (for example, to register with
    dependency injection tool which will instantiate it for you). If you just want an a
    client instance to use directly, use the akismet_async_client fixture instead.

    By default, this will return a class whose verify_key() method always returns True,
    and whose comment_check() method always returns akismet.CheckResponse.SPAM. To
    change this, use the pytest mark akismet_client on your test, and pass either or
    both of the following keyword arguments:

    * comment_check_response: an akismet.CheckResponse which will be used as the return
      value of the comment_check() method.

    * verify_key_response: a bool which will be used as the return value of the
      verify_key() method.

    """
    _comment_check_response, _verify_key_response = _get_mark_arguments(request)

    class _AsyncClient(TestAsyncClient):
        """
        Test Akismet client.

        """

        comment_check_response = _comment_check_response
        verify_key_response = _verify_key_response

    return _AsyncClient


@pytest.fixture
def akismet_sync_client(akismet_sync_class: type[SyncClient]) -> SyncClient:
    """
    A synchronous Akismet test client instance.

    Use this fixture if you just need a client instance to usedirectly. If you instead
    need the client class (for example, to register with a dependency injection system
    which will instantiate it for you), use the akismet_sync_class fixture instead.

    By default, this will return a client whose verify_key() method always returns True,
    and whose comment_check() method always returns akismet.CheckResponse.SPAM. To change
    this, use the pytest mark akismet_client on your test, and pass either or both of the
    following keyword arguments:

    * comment_check_response: an akismet.CheckResponse which will be used as the return
      value of the comment_check() method.

    * verify_key_response: a bool which will be used as the return value of the
      verify_key() method.

    """
    return akismet_sync_class()


@pytest.fixture
def akismet_async_client(akismet_async_class: type[AsyncClient]) -> AsyncClient:
    """
    An asynchronous Akismet test client instance.

    Use this fixture if you just need a client instance to usedirectly. If you instead
    need the client class (for example, to register with a dependency injection system
    which will instantiate it for you), use the akismet_async_class fixture instead.

    By default, this will return a client whose verify_key() method always returns True,
    and whose comment_check() method always returns akismet.CheckResponse.SPAM. To change
    this, use the pytest mark akismet_client on your test, and pass either or both of the
    following keyword arguments:

    * comment_check_response: an akismet.CheckResponse which will be used as the return
      value of the comment_check() method.

    * verify_key_response: a bool which will be used as the return value of the
      verify_key() method.
    """
    return akismet_async_class()


@pytest.fixture
def akismet_ham_role() -> str:  # pragma: no cover
    """
    Return a user_role value which Akismet always considers to be non-spam.

    """
    return "administrator"


@pytest.fixture
def akismet_spam_author() -> str:  # pragma: no cover
    """
    Return a comment_author value which Akismet always considers to be spam.

    """
    return "akismet-guaranteed-spam"


def pytest_configure(config: pytest.Config):
    """
    Register custom markers.

    """
    config.addinivalue_line(
        "markers",
        "akismet_client: Configure the Akismet test client fixtures.",
    )



"""
Tests for the sync test client.

"""

import pytest

import akismet

pytestmark = [pytest.mark.anyio, pytest.mark.async_client, pytest.mark.test_client]


async def test_comment_check_default_not_spam(
    akismet_config: akismet.Config, akismet_common_kwargs: dict
):
    """
    The async default configuration of the test client marks all content as non-spam.

    """
    client = akismet.TestAsyncClient(config=akismet_config)
    assert (
        await client.comment_check(comment_content="Test", **akismet_common_kwargs)
        == akismet.CheckResponse.HAM
    )


@pytest.mark.parametrize("response_value", list(akismet.CheckResponse))
async def test_comment_check_configuration(
    akismet_config: akismet.Config,
    akismet_common_kwargs: dict,
    response_value: akismet.CheckResponse,
):
    """
    Setting comment_check_response explicitly will cause the test client to mark content
    accordingly.

    """

    class _Client(akismet.TestAsyncClient):
        """
        Test client with explicit comment_check_response.

        """

        comment_check_response = response_value

    client = _Client(config=akismet_config)
    assert (
        await client.comment_check(comment_content="Test", **akismet_common_kwargs)
        == response_value
    )


async def test_verify_key_default(akismet_config: akismet.Config):
    """
    The async default configuration of the test client succeeds at key verification.

    """
    client = akismet.TestAsyncClient(config=akismet_config)
    assert await client.verify_key()


async def test_verify_key_explicit_success(akismet_config: akismet.Config):
    """
    Setting verify_key_response explicitly to True will cause the test client to succeed
    at key verification.

    """

    class _Client(akismet.TestAsyncClient):
        """
        Test client with explicit verify_key_response.

        """

        verify_key_response = True

    # Explicit configuration succeeds.
    client = _Client(config=akismet_config)
    assert await client.verify_key()

    # Implicit configuration succeeds.
    await _Client.validated_client()


async def test_verify_key_explicit_failure(akismet_config: akismet.Config):
    """
    Setting verify_key_response explicitly to False will cause the test client to fail
    at key verification.

    """

    class _Client(akismet.TestAsyncClient):
        """
        Test client with explicit verify_key_response.

        """

        verify_key_response = False

    # Explicit configuration fails.
    client = _Client(config=akismet_config)
    assert not await client.verify_key()

    # Implicit configuration fails.
    with pytest.raises(akismet.APIKeyError):
        await _Client.validated_client()



"""
End-to-end (live requests to the Akismet web service) tests for the async client.

"""

import pytest

import akismet
from akismet import _common

pytestmark = [pytest.mark.anyio, pytest.mark.async_client, pytest.mark.end_to_end]


async def test_construct_config_valid():
    """
    With a valid configuration, constructing a client succeeds.

    """
    await akismet.AsyncClient.validated_client()


async def test_construct_config_invalid_key(
    monkeypatch: pytest.MonkeyPatch, akismet_bad_config: akismet.Config
):
    """
    With an invalid API key, constructing a client raises an APIKeyError.

    """
    monkeypatch.setenv(_common._KEY_ENV_VAR, akismet_bad_config.key)
    monkeypatch.setenv(_common._URL_ENV_VAR, akismet_bad_config.url)
    with pytest.raises(akismet.APIKeyError):
        await akismet.AsyncClient.validated_client()


async def test_verify_key_valid():
    """
    verify_key() returns True when the config is valid.

    """
    client = akismet.AsyncClient()
    assert await client.verify_key()


async def test_verify_key_invalid(akismet_bad_config: akismet.Config):
    """
    verify_key() returns False when the config is invalid.

    """
    client = akismet.AsyncClient()
    assert not await client.verify_key(akismet_bad_config.key, akismet_bad_config.url)


@pytest.mark.parametrize(
    ["method_name", "pass_args"],
    [
        ("comment_check", True),
        ("submit_ham", True),
        ("submit_spam", True),
        ("key_sites", False),
        ("usage_limit", False),
    ],
    ids=["comment_check", "submit_nam", "submit_spam", "key_sites", "usage_limit"],
)
async def test_request_with_invalid_key(
    akismet_bad_config: akismet.Config,
    akismet_end_to_end_kwargs: dict,
    method_name: str,
    pass_args: bool,
):
    """
    The request methods other than verify_key() raise akismet.APIKeyError if called with
    an invalid API key/URL.

    """
    client = akismet.AsyncClient(config=akismet_bad_config)
    method = getattr(client, method_name)
    args = akismet_end_to_end_kwargs if pass_args else {}
    with pytest.raises(akismet.APIKeyError):
        await method(**args)


async def test_comment_check_spam(
    akismet_end_to_end_kwargs: dict, akismet_spam_author: str
):
    """
    comment_check() returns the SPAM value when Akismet declares the content to be spam.

    """
    client = akismet.AsyncClient()
    assert (
        await client.comment_check(
            comment_content="test",
            comment_author=akismet_spam_author,
            **akismet_end_to_end_kwargs,
        )
        == akismet.CheckResponse.SPAM
    )


async def test_comment_check_ham(
    akismet_end_to_end_kwargs: dict, akismet_ham_role: str
):
    """
    comment_check() returns the HAM value when Akismet declares the content to be ham.

    """
    client = akismet.AsyncClient()
    assert (
        await client.comment_check(
            comment_content="test",
            user_role=akismet_ham_role,
            **akismet_end_to_end_kwargs,
        )
        == akismet.CheckResponse.HAM
    )


async def test_submit_ham(akismet_end_to_end_kwargs: dict, akismet_ham_role: str):
    """
    submit_ham() returns True when Akismet accepts the submission.

    """
    client = akismet.AsyncClient()
    assert await client.submit_ham(
        comment_content="test", user_role=akismet_ham_role, **akismet_end_to_end_kwargs
    )


async def test_submit_spam(akismet_end_to_end_kwargs: dict, akismet_spam_author: str):
    """
    submit_spam() returns True when Akismet accepts the submission.

    """
    client = akismet.AsyncClient()
    assert await client.submit_spam(
        comment_content="test",
        comment_author=akismet_spam_author,
        **akismet_end_to_end_kwargs,
    )



"""
Tests for the async client's error behaviors.

"""

from http import HTTPStatus

import httpx
import pytest

import akismet

pytestmark = [pytest.mark.anyio, pytest.mark.async_client, pytest.mark.constructors]

check_all_methods = pytest.mark.parametrize(
    ["method_name", "pass_args"],
    [
        ("comment_check", True),
        ("key_sites", False),
        ("submit_ham", True),
        ("submit_spam", True),
        ("usage_limit", False),
        ("verify_key", False),
    ],
    ids=[
        "comment_check",
        "key_sites",
        "submit_ham",
        "submit_spam",
        "usage_limit",
        "verify_key",
    ],
)


@check_all_methods
@pytest.mark.parametrize(
    "status_code",
    [
        pytest.param(
            code.value, marks=pytest.mark.akismet_fixed_response(status_code=code)
        )
        for code in HTTPStatus
        if 400 <= code < 600
    ],
    ids=[code.value for code in HTTPStatus if 400 <= code < 600],
)
async def test_error_status(
    akismet_async_client_fixed_response: akismet.AsyncClient,
    akismet_common_kwargs: dict,
    status_code: int,
    method_name: str,
    pass_args: bool,
):
    """
    RequestError is raised when a POST request to Akismet responds with an HTTP status
    code indicating an error.

    """
    method = getattr(akismet_async_client_fixed_response, method_name)
    args = akismet_common_kwargs if pass_args else {}
    with pytest.raises(
        akismet.RequestError,
        match=f"Akismet responded with error status: {status_code}",
    ):
        await method(**args)


@check_all_methods
@pytest.mark.akismet_exception_response(exception_class=httpx.TimeoutException)
async def test_error_timeout(
    akismet_async_client_exception: akismet.AsyncClient,
    akismet_common_kwargs: dict,
    method_name: str,
    pass_args: bool,
):
    """
    RequestError is raised when the request to Akismet times out.

    """
    method = getattr(akismet_async_client_exception, method_name)
    args = akismet_common_kwargs if pass_args else {}
    with pytest.raises(akismet.RequestError, match="Akismet timed out."):
        await method(**args)


@check_all_methods
@pytest.mark.akismet_exception_response(exception_class=httpx.RequestError)
async def test_error_other_httpx(
    akismet_async_client_exception: akismet.AsyncClient,
    akismet_common_kwargs: dict,
    method_name: str,
    pass_args: bool,
):
    """
    RequestError is raised when a generic httpx request error occurs.

    """
    method = getattr(akismet_async_client_exception, method_name)
    args = akismet_common_kwargs if pass_args else {}
    with pytest.raises(akismet.RequestError, match="Error making request to Akismet."):
        await method(**args)


@check_all_methods
@pytest.mark.akismet_exception_response(exception_class=TypeError)
async def test_error_other(
    akismet_async_client_exception: akismet.AsyncClient,
    akismet_common_kwargs: dict,
    method_name: str,
    pass_args: bool,
):
    """
    RequestError is raised when any other (non-httpx) exception occurs during the
    request.

    """
    method = getattr(akismet_async_client_exception, method_name)
    args = akismet_common_kwargs if pass_args else {}
    with pytest.raises(akismet.RequestError, match="Error making request to Akismet."):
        await method(**args)


@pytest.mark.parametrize("method_name", ["comment_check", "submit_ham", "submit_spam"])
async def test_unknown_argument(
    akismet_async_client: akismet.AsyncClient,
    akismet_common_kwargs: dict,
    method_name: str,
):
    """
    UnknownArgumentError is raised when an argument outside the supported set is passed
    to one of the POST request methods.

    """
    with pytest.raises(akismet.UnknownArgumentError, match="bad_argument"):
        await getattr(akismet_async_client, method_name)(
            bad_argument=1, **akismet_common_kwargs
        )


@pytest.mark.parametrize(
    ["method_name", "pass_args"],
    [
        ("comment_check", True),
        ("submit_ham", True),
        ("submit_spam", True),
        ("verify_key", False),
    ],
    ids=["comment_check", "submit_ham", "submit_spam", "verify_key"],
)
@pytest.mark.akismet_fixed_response(response_text="bad")
async def test_protocol_error(
    akismet_async_client_fixed_response: akismet.AsyncClient,
    akismet_common_kwargs: dict,
    method_name: str,
    pass_args: bool,
):
    """
    ProtocolError is raised when ``comment_check()`` receives an unexpected
    response.

    """
    method = getattr(akismet_async_client_fixed_response, method_name)
    args = akismet_common_kwargs if pass_args else {}
    with pytest.raises(akismet.ProtocolError):
        await method(**args)
